"""
    Query Tools
    ===== =====

_RelationalJoinsMixin
    Methods to allow specifying joins.

_RelationalLogicMixin
    Methods to perform relational type joins.
"""
# $Id: QueryTools.py,v 1.1.1.1 2002/02/03 17:07:16 root Exp $
#
# $Log: QueryTools.py,v $
# Revision 1.1.1.1  2002/02/03 17:07:16  root
#
#
# Revision 1.2  2002/01/21 00:58:08  adrian
# Rebased all queries and datacombiner on the Aqueduct DA
# Added argument support
# Added Drill-Down support to ZReportTool
#
# Revision 1.1  2002/01/15 20:26:04  adrian
# Factored out common code in all-QT and ZDC
# Added support for comparing against constants in all-QT
# Added support for multi-column joins in all-QT
# Added sorting to ZRT
# Added no-print group type to ZRT
# You now set the connection ID when creating a ZVQ
# added property support to ZVQ
# Corrected display of calculated columns in ZRT
#
#
__version__ = '$Revision: 1.1.1.1 $'[11:-2]

import Globals
import string

def _getTableName(columnName):
    """ Split the table name from a FQ column name """
    if string.find(columnName, '.')==-1:
        return None
    return columnName[:string.find(columnName, '.')]

def _getColumnName(columnName):
    """ Split the column name from a FQ column name """
    if string.find(columnName, '.')==-1:
        return columnName
    return columnName[string.find(columnName, '.')+1:]


class _RelationalJoinsMixin:
    """ Add/Remove Joins """
    joins = []
    manage_Joins = Globals.HTMLFile('www/queryJoins', globals())
    manage_options = (
        (
            {'label': 'Joins', 'action': 'manage_Joins',
             'help': ('ZDataQueryKit', 'QueryJoins.stx'), },
        )
    )
    relationActions = (
        '=',
        '!=',
        '>',
        '<',
        '>=',
        '<=',
        '=*',
        '*=',
    )

    def _RJM_Reset(self):
        """ Reset the joins in the mix """
        self.joins = []

    def _RJM_RemoveTable(self, table):
        """ Remove all joins that reference this table or parameter """
        joins = []
        for item in self.joins:
            if table not in (
                    _getTableName(item['leftColumn']),
                    _getTableName(item['rightColumn']),
                    item['leftColumn'],
                    item['rightColumn']
                ):
                joins.append(item)
        self.joins = joins

    def _RJM_RemoveInvalidJoins(self, columnCache):
        """ Remove joins referencing tables not in columnCache """

    def manage_addJoin(self, leftColumn, leftValue, relation, rightColumn, rightValue, REQUEST=None):
        """ Add a new table join """
        joins = self.joins
        if leftColumn == '-':
            leftColumn = leftValue
        if rightColumn == '-':
            rightColumn = rightValue
        joins.append({'leftColumn': str(leftColumn), 'relation': str(relation), 'rightColumn': str(rightColumn)})
        self.joins = joins
        self._updateTemplate()
        if REQUEST:
            message = "Join created."
            return self.manage_Joins(self, REQUEST, manage_tabs_message=message)

    def manage_delJoin(self, REQUEST=None):
        """ Delete one or more table joins """
        joins = self.joins
        for item in range(len(self.joins)-1, -1, -1):
            if REQUEST.get('dl_%d' % item):
                del joins[item]
        self.joins = joins
        self._updateTemplate()
        if REQUEST:
            message = "Join(s) deleted."
            return self.manage_Joins(self, REQUEST, manage_tabs_message=message)

class _RelationalLogicMixin(
        _RelationalJoinsMixin,
    ):
    """ Perform joins on data """
    relationActions = (
        '=',
        '!=',
        '>',
        '<',
        '>=',
        '<=',
        '=*',
        '*=',
        '*=*',
    )

    def _combine_rows(self, rowL, prefixL, rowR, prefixR):
        """ Combine the fields from rowL with
            those from rowR. """
        out = {}
        for fieldL in rowL.keys():
            if prefixL:
                out["%s.%s" % (prefixL, fieldL)] = rowL[fieldL]
            else:
                out[fieldL] = rowL[fieldL]
        for fieldR in rowR.keys():
            if prefixR:
                out["%s.%s" % (prefixR, fieldR)] = rowR[fieldR]
            else:
                out[fieldR] = rowR[fieldR]
        return out

    def _product(self, dsL, prefixL, dsR, prefixR, joins=[]):
        """ Generate the product of dsL with dsR
            adding to the output only those rows that
            match the joins. DSL is either the table
            mentioned in joins[n]['leftColumn'] or
            an existing product the contains it. """
        out = []
        for rowL in dsL:
            for rowR in dsR:
                row = self._combine_rows(rowL, prefixL, rowR, prefixR)
# Test join constraints
                include = 1
                for join in joins:
                    leftValue = row[join['leftColumn']]
                    rightValue = row[join['rightColumn']]
                    rel = join['relation']
                    if rel == '=':
                        if not leftValue == rightValue: include = 0
                    elif rel == '!=':
                        if not leftValue != rightValue: include = 0
                    elif rel == '>':
                        if not leftValue > rightValue: include = 0
                    elif rel == '<':
                        if not leftValue < rightValue: include = 0
                    elif rel == '>=':
                        if not leftValue >= rightValue: include = 0
                    elif rel == '<=':
                        if not leftValue <= rightValue: include = 0
                    else:
                        raise "Invalid Relational Operator"
                if include:
                    out.append(row)
        return out

    def _match_joins(self, dsList1, dsList2):
        """ find all the joins referencing something
            from list1 and something from list2 """
        out = []
        for join in self.joins:
            dsL = _getTableName(join['leftColumn'])
            dsR = _getTableName(join['rightColumn'])
            if (dsL in dsList1 and dsR in dsList2) or (dsL in dsList2 and dsR in dsList1):
                out.append(join)
        return out
